#!/usr/bin/perl
use DBI;
use Time::HiRes qw(time);


open(my $fh, '<', $ENV{DBI_CREDENTIAL_FILE}) or die;
$_ = <$fh>;
my ($conn, $user, $pw) = split;
my $dbh = DBI->connect($conn, $user, $pw, { RaiseError => 1 });

cleanup();
bench(\&create_heap_ix_person_id);
bench(\&create_heap_ix_person_id_doc_name);
bench(\&create_heap_ix_person_id_bitmap);
bench(\&create_iot);
bench(\&create_iot_compress);
bench(\&create_cluster);

sub bench {
    my ($create) = @_;
    $create->();
    print "=" x 72, "\n\n";
    fill_table();
    gather_stats();
    explain();
    for (1 .. 5) {
	my $t0 = time();
	my $rs = $dbh->selectall_arrayref("SELECT doc_name FROM t WHERE person_id = 221");
	print time() - $t0, " seconds\n";
    }
    cleanup();
}

sub create_heap_ix_person_id {
    print "\n\nHeap organized, index on person_id\n";
    $dbh->do("
		CREATE TABLE t (
		doc_id    VARCHAR2(10),
		person_id NUMBER(7),
		doc_name  VARCHAR2(45))
	    ");

    $dbh->do("
		CREATE INDEX ix_t_person_id
		ON t(person_id)
	    ");
}

sub create_heap_ix_person_id_doc_name {
    print "\n\nHeap organized, index on person_id, doc_name\n";
    $dbh->do("
		CREATE TABLE t (
		doc_id    VARCHAR2(10),
		person_id NUMBER(7),
		doc_name  VARCHAR2(45))
	    ");

    $dbh->do("
		CREATE INDEX ix_t_person_id
		ON t(person_id, doc_name)
	    ");
}

sub create_heap_ix_person_id_bitmap {
    print "\n\nHeap organized, bitmap index on person_id\n";
    $dbh->do("
		CREATE TABLE t (
		doc_id    VARCHAR2(10),
		person_id NUMBER(7),
		doc_name  VARCHAR2(45))
	    ");

    $dbh->do("
		CREATE bitmap INDEX ix_t_person_id
		ON t(person_id)
	    ");
}

sub create_iot {
    print "\n\nIndex organized on person_id, doc_id\n";
    $dbh->do("
		CREATE TABLE t (
		    doc_id    VARCHAR2(10),
		    person_id NUMBER(7),
		    doc_name  VARCHAR2(45),
		    constraint t_pk primary key(person_id, doc_id)
		)
		organization index
	    ");
}

sub create_iot_compress {
    print "\n\nIndex organized on person_id, doc_id; compressed\n";
    $dbh->do("
		CREATE TABLE t (
		    doc_id    VARCHAR2(10),
		    person_id NUMBER(7),
		    doc_name  VARCHAR2(45),
		    constraint t_pk primary key(person_id, doc_id)
		)
		organization index compress
	    ");
}

sub create_cluster {
    print "\n\nSingle table cluster hashkeys 16\n";
    $dbh->do("
		CREATE cluster cl_t (
		    person_id NUMBER(7)
		)
		    single table 
		    hashkeys 16
		    hash is (mod(person_id, 16))
	     ");
    $dbh->do("create table t (
		    doc_id    VARCHAR2(10),
		    person_id NUMBER(7),
		    doc_name  VARCHAR2(45)
		)
		cluster cl_t (person_id)
	    ");
}

sub fill_table {

    $dbh->do("
		CREATE SEQUENCE seq_t
	    ");

    $dbh->do(q{
		DECLARE
		  t_docname dbms_sql.VARCHAR2_TABLE;

		CURSOR c IS
		  SELECT object_name
		  FROM dba_objects;
		BEGIN
		   OPEN c;
		   LOOP
		     FETCH c BULK COLLECT INTO t_docname LIMIT 250;

		     FORALL i IN 1..t_docname.COUNT
		     INSERT INTO t
		     (doc_id, person_id, doc_name)
		     VALUES
		     (seq_t.NEXTVAL, MOD(seq_t.CURRVAL, 233), t_docname(i));

		     EXIT WHEN c%NOTFOUND;
		   END LOOP;
		   COMMIT;
		   CLOSE c;
		END;
	    });
}

sub gather_stats {
    $dbh->do(q{
		begin
		    dbms_stats.gather_table_stats(USER, 'T', CASCADE=>TRUE);
		end;
	    });
}

sub cleanup {
    local $dbh->{RaiseError};
    local $dbh->{PrintError};
    $dbh->do("drop table t");
    $dbh->do("drop cluster cl_t");
    $dbh->do("drop sequence seq_t");
}

sub explain {
    $dbh->do("explain plan for SELECT doc_name FROM t WHERE person_id = 221");
    my $rs = $dbh->selectall_arrayref("SELECT * FROM table(dbms_xplan.display)");
    for my $r (@$rs) {
	for my $c (@$r) {
	    print "$c\n";
	}
    }
}
